from solaredge2mqtt.models.base import Component, EnumModel
from solaredge2mqtt.models.energy import Energy, EnergyPeriod, EnergyQuery
from solaredge2mqtt.models.forecast import (
    EnergyForecast,
    Forecast,
    ForecastAccount,
    ForecastAPIKeyInfo,
)
from solaredge2mqtt.models.modbus import SunSpecBattery, SunSpecInverter, SunSpecMeter
from solaredge2mqtt.models.monitoring import (
    LogicalInfo,
    LogicalInverter,
    LogicalModule,
    LogicalString,
)
from solaredge2mqtt.models.powerflow import Powerflow
from solaredge2mqtt.models.wallbox import WallboxAPI



from datetime import datetime, timedelta, timezone

from influxdb_client import Point

from solaredge2mqtt.models.base import EnumModel, Solaredge2MQTTBaseModel


class ForecastAccount(EnumModel):
    PUBLIC = "Public", 1, 600, 60
    PERSONAL = "Personal", 1, 600, 30
    PERSONAL_PLUS = "Personal Plus", 2, 300, 15
    PROFESSSIONAL = "Professional", 3, 300, 15
    PROFESSSIONAL_PLUS = "Professional Plus", 4, 300, 15

    def __init__(
        self,
        account_name: str,
        allowed_strings: int,
        interval_in_seconds: int,
        resolution_in_minutes: int,
    ) -> None:
        self._account_name = account_name
        self._allowed_strings = allowed_strings
        self._interval_in_seconds = interval_in_seconds
        self._resolution_in_minutes = resolution_in_minutes

    @property
    def account_name(self) -> str:
        return self.account_name

    @property
    def allowed_strings(self) -> int:
        return self._allowed_strings

    @property
    def interval_in_seconds(self) -> int:
        return self._interval_in_seconds

    @property
    def resolution_in_minutes(self) -> int:
        return self._resolution_in_minutes


class ForecastAPIKeyInfo(Solaredge2MQTTBaseModel):
    account: ForecastAccount = ForecastAccount.PUBLIC

    model_config = {"extra": "ignore"}


class Forecast(Solaredge2MQTTBaseModel):
    watts: dict[datetime, int]
    watt_hours_period: dict[datetime, int]
    watt_hours: dict[datetime, int]
    watt_hours_day: dict[str, int]

    @property
    def influxdb_points(self) -> list[Point]:
        merged_power_and_energy = {
            k: (self.watts.get(k, 0), self.watt_hours_period.get(k, 0))
            for k in set(self.watts) | set(self.watt_hours_period)
        }

        points = [
            Point("forecast")
            .field("power", power)
            .field("energy", energy / 1000)
            .time(timestamp.astimezone(timezone.utc), write_precision="s")
            for timestamp, (power, energy) in merged_power_and_energy.items()
        ]

        return points


class EnergyForecast(Solaredge2MQTTBaseModel):
    current_day: int
    next_day: int
    current_hour: int
    next_hour: int
    next_three_hours: int

    def __init__(self, forecast: Forecast):
        time_current_hour = datetime.now().replace(minute=0, second=0, microsecond=0)
        time_next_hour = time_current_hour + timedelta(hours=1)

        current_hour = self._calc_energy(forecast, time_current_hour, time_next_hour)

        time_next_two_hours = time_current_hour + timedelta(hours=2)

        next_hour = self._calc_energy(forecast, time_next_hour, time_next_two_hours)

        time_next_three_hours = time_current_hour + timedelta(hours=3)

        next_three_hours = self._calc_energy(
            forecast, time_current_hour, time_next_three_hours
        )

        time_current_day = datetime.now().replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        time_next_day = time_current_day + timedelta(days=1)
        current_day = self._calc_energy(forecast, time_current_day, time_next_day)

        time_next_two_days = time_current_day + timedelta(days=2)
        next_day = self._calc_energy(forecast, time_next_day, time_next_two_days)

        super().__init__(
            current_day=current_day,
            next_day=next_day,
            current_hour=current_hour,
            next_hour=next_hour,
            next_three_hours=next_three_hours,
        )

    @staticmethod
    def _calc_energy(
        forecast: Forecast, start_time: datetime, end_time: datetime
    ) -> int:
        energy = 0
        for time, watt_hours in forecast.watt_hours_period.items():
            if time >= start_time and time < end_time:
                energy += watt_hours
            elif time >= end_time:
                break
        return energy



import asyncio as aio
from typing import Optional

from solaredge2mqtt.exceptions import InvalidDataException
from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import EnergyPeriod, EnergyQuery, Powerflow
from solaredge2mqtt.mqtt import MQTTClient
from solaredge2mqtt.persistence.influxdb import InfluxDB
from solaredge2mqtt.service.modbus import Modbus
from solaredge2mqtt.service.wallbox import WallboxClient
from solaredge2mqtt.settings import ServiceSettings


class BaseLoops:
    def __init__(
        self,
        settings: ServiceSettings,
        mqtt: MQTTClient,
        influxdb: Optional[InfluxDB] = None,
    ):
        self.settings = settings

        self.mqtt = mqtt

        self.influxdb = influxdb

        self.modbus = Modbus(self.settings.modbus)

        self.wallbox = (
            WallboxClient(self.settings.wallbox)
            if self.settings.is_wallbox_configured
            else None
        )

    async def powerflow_loop(self):
        results = await aio.gather(
            self.modbus.loop(),
            self.wallbox.loop()
            if self.settings.is_wallbox_configured
            else aio.sleep(0),
        )

        inverter_data, meters_data, batteries_data = results[0]

        if any(data is None for data in [inverter_data, meters_data, batteries_data]):
            raise InvalidDataException("Invalid modbus data")

        for battery in batteries_data.values():
            if not battery.is_valid:
                logger.debug(battery)
                raise InvalidDataException("Invalid battery data")

        evcharger = 0

        if self.settings.is_wallbox_configured:
            wallbox_data = results[1]
            logger.trace("Wallbox: {wallbox_data.power} W", wallbox_data=wallbox_data)
            evcharger = wallbox_data.power

        powerflow = Powerflow(inverter_data, meters_data, batteries_data, evcharger)
        if not powerflow.is_valid:
            logger.info(powerflow)
            raise InvalidDataException("Invalid powerflow data")

        if Powerflow.is_not_valid_with_last(powerflow):
            logger.debug(powerflow)
            raise InvalidDataException("Value change not valid, skipping this loop")

        logger.debug(powerflow)
        logger.info(
            "Powerflow: PV {pv_production} W, Inverter {inverter.power} W, "
            + "House {consumer.house} W, Grid {grid.power} W, "
            + "Battery {battery.power} W, Wallbox {consumer.evcharger} W",
            pv_production=powerflow.pv_production,
            inverter=powerflow.inverter,
            consumer=powerflow.consumer,
            grid=powerflow.grid,
            battery=powerflow.battery,
        )

        await self.mqtt.publish_to(inverter_data.mqtt_topic, inverter_data)
        for key, component in {**meters_data, **batteries_data}.items():
            await self.mqtt.publish_to(
                f"{component.mqtt_topic}/{key.lower()}", component
            )

        if wallbox_data is not None:
            await self.mqtt.publish_to(wallbox_data.mqtt_topic, wallbox_data)

        await self.mqtt.publish_to("powerflow", powerflow)

        if self.influxdb is not None:
            self.influxdb.write_components(
                inverter_data, meters_data, batteries_data, wallbox_data
            )
            self.influxdb.write_powerflow(powerflow)

            self.influxdb.flush_loop()

    async def energy_loop(self):
        for period in EnergyPeriod:
            energy = self.influxdb.query_energy(period)
            if energy is None:
                if period.query == EnergyQuery.LAST:
                    logger.info(
                        "No data found for {period}, skipping this loop", period=period
                    )
                else:
                    raise InvalidDataException(f"No energy data for {period}")

                continue

            logger.info(
                "Read from influxdb {period} energy: {energy.pv_production} kWh",
                period=period,
                energy=energy,
            )

            await self.mqtt.publish_to(f"energy/{period.topic}", energy)



from typing import Optional

from requests.exceptions import HTTPError

from solaredge2mqtt.exceptions import ConfigurationException
from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import (
    EnergyForecast,
    Forecast,
    ForecastAccount,
    ForecastAPIKeyInfo,
)
from solaredge2mqtt.mqtt import MQTTClient
from solaredge2mqtt.persistence.influxdb import InfluxDB
from solaredge2mqtt.service.http import HTTPClient
from solaredge2mqtt.settings import ForecastSettings

LOCATION_PART = "/{latitude}/{longitude}"

API_KEY_INFO = "https://api.forecast.solar/{api_key}/info"

CHECK_URL = "https://api.forecast.solar/check" + LOCATION_PART
ESTIMATE_URL = "https://api.forecast.solar/{api_key}/estimate" + LOCATION_PART


class ForecastAPI(HTTPClient):

    def __init__(
        self,
        settings: ForecastSettings,
        mqtt: MQTTClient,
        influxdb: Optional[InfluxDB] = None,
    ) -> None:
        super().__init__("Forecast API")
        self.settings = settings

        self.mqtt = mqtt

        self.influxdb = influxdb

        self.account = self.get_account()

        self._check_location()

    def get_account(self) -> ForecastAccount:
        account = ForecastAccount.PUBLIC
        if self.settings.api_key is not None:
            try:
                result = self._get(
                    API_KEY_INFO.format(
                        api_key=self.settings.api_key.get_secret_value()
                    )
                )
                api_key_info = ForecastAPIKeyInfo(**result["result"])
                account = api_key_info.account
            except HTTPError as error:
                raise ConfigurationException(
                    "Cannot get forecast account info"
                ) from error

        logger.info(
            "Forecast account: {account} (max. {account.allowed_strings} strings)",
            account=account,
        )

        if account.allowed_strings < 2 and self.settings.string2 is not None:
            raise ConfigurationException(
                "The configured API key does not support two strings."
            )

        return account

    def _check_location(self) -> None:
        try:
            result = self._get(
                CHECK_URL.format(
                    latitude=self.settings.latitude, longitude=self.settings.longitude
                )
            )

            logger.info(
                "Forecast configuration for location: {place}",
                place=result["result"]["place"],
            )
        except HTTPError as error:
            raise ConfigurationException(
                "Cannot check location for forecast"
            ) from error

    async def loop(self) -> Forecast | None:
        forecast = None

        try:
            logger.info("Read forecast from API")
            result = self._get(
                self.estimate_url, headers={"Accept": "application/json"}
            )

            forecast = Forecast(**result["result"])
            logger.debug(forecast.model_dump_json(indent=4))

            energy = EnergyForecast(forecast)

            logger.info("Read forecast from API: {energy}", energy=energy)

            await self.mqtt.publish_to("forecast/energy", energy)

            if self.influxdb:
                logger.info("Write forecast to influxdb")
                self.influxdb.write_points_to_aggregated_bucket(
                    forecast.influxdb_points
                )

        except HTTPError as error:
            logger.warning("Cannot get forecast: {error}", error=error)

        return forecast

    @property
    def estimate_url(self) -> str:
        estimate_url = ESTIMATE_URL.format(
            api_key=self.settings.api_key.get_secret_value(),
            latitude=self.settings.latitude,
            longitude=self.settings.longitude,
        )

        estimate_url += self.settings.string1.url_string

        if self.settings.string2 is not None:
            estimate_url += self.settings.string2.url_string

        return estimate_url
